import traceback
import threading
from typing import List
from datetime import datetime

from google.cloud import firestore_v1
from firebase_admin import firestore

from boastlabs.functions.retriable.execution.config import Status
from boastlabs.functions.retriable.execution.progress import Progress
from boastlabs.functions.retriable.execution.events.handling import EventWorkDone


class State(object):

    created_at: datetime
    modified_at: datetime

    active: bool
    status: str

    _progress: Progress

    def __init__(self, service_name, db, job_ref):
        self.service_name = service_name
        self.db = db

        self.job_doc = job_ref
        self.etl_job_doc = self.job_doc.parent.parent
        self.fiscal_year_doc = self.etl_job_doc.parent.parent
        self.provider_doc = self.fiscal_year_doc.parent.parent

        self.job_dict = self.job_doc.get().to_dict()
        self.etl_job_dict = self.etl_job_doc.get().to_dict()
        self.provider_dict = self.provider_doc.get().to_dict()
        self.fiscal_year_dict = self.fiscal_year_doc.get().to_dict()

        self.active = self.job_dict.get('active', False)
        self.status = self.job_dict.get('status', Status.NEW)
        self.created_at = self.job_dict.get('created_at', None)
        self.modified_at = self.job_dict.get('modified_at', None)

        self.fy_start_date = datetime.strptime(self.fiscal_year_dict['start_date'], '%Y-%m-%d')
        self.fy_end_date = datetime.strptime(self.fiscal_year_dict['end_date'], '%Y-%m-%d')

        self._read_lock = threading.RLock()
        self._write_lock = threading.RLock()

        # Load progress
        if self.job_doc.collection('progress').document('progress').get().exists:
            self._progress = Progress(
                id='progress',
                progress=self.job_doc.collection('progress').document('progress').get().to_dict().get('progress', None))
        else:
            self._progress = Progress(id='progress', progress=None)

        # print('# Progress', json.dumps(self._progress.to_dict()))

    def set_created_at(self):
        if self.created_at is None:
            self.created_at = self.job_doc.update({'created_at': firestore.SERVER_TIMESTAMP})

    def set_modified_at(self):
        self.modified_at = self.job_doc.update({'modified_at': firestore.SERVER_TIMESTAMP})
        self.etl_job_doc.update({'modified_at': firestore.SERVER_TIMESTAMP})

    def set_active(self, active: bool):
        self.active = active
        self.job_doc.update({'active': active})

    def set_status(self, status: str, error=None):
        if status == Status.DONE:
            self.notify_work_done()

            # Save the status last if all worked well
            self.status = Status.DONE
            self.active = False

        elif status == Status.FAILED:
            # Save the status first regardless if all else will fail
            self.status = Status.FAILED
            self.active = False

            error_string = repr(error)
            error_stack = traceback.format_exc()

            self.job_doc.update({'active': False, 'error': error_string, 'error_stack': error_stack})
            self.etl_job_doc.update({'status': status, 'error': error_string, 'error_stack': error_stack})

        else:
            self.job_doc.update({'status': status})
            self.etl_job_doc.update({'status': status})
            self.status = status

    def get_last_execution_date(self):
        """Gets last executed date from previous jobs.
        Query files in descending by created_at and get the last executed time.

        Returns:
            last_executed_date:datetime object
        """
        query = self.etl_job_doc.parent.order_by(u'created_at', direction=firestore.Query.DESCENDING).limit(2)
        for result in query.stream():
            # Exclude the current job ID and include previous job
            if result.id != self.etl_job_doc.id:
                return result.to_dict()['created_at']
        return None

    def save_progress(self):
        with self._write_lock:
            self.set_modified_at()
            self.job_doc.collection('progress').document('progress').set(
                {'progress': self._progress.to_dict()},
                merge=True
            )

    def get_progress(self, items: List[str]) -> Progress:
        """
        todo: add documentation, mandatory function
        :param items:
        :return:
        """

        # This will basically do a return progress[user_name][organisation][repo][kind]

        with self._read_lock:
            p = self._progress
            for item in items:
                p = p[item]
            return p

    def notify_work_done(self):
        event_creator = EventWorkDone(db=self.db)
        event_creator.create(
            job_ref=self.job_doc,
            parent_ref=self.etl_job_doc,
            service_name=self.service_name,
            service_status=Status.DONE)



import threading
import logging
from boastlabs.functions.retriable.execution.exceptions import TimeoutException


class TimeoutEvent(threading.Event):
    pass


class Timer:
    _logger: logging.Logger

    def __init__(self, event_id):
        self.event_id = event_id
        self.timeout_event = TimeoutEvent()
        self.logger = self._init_logger()

    def _init_logger(self):
        logger = logging.getLogger(f"[{self.event_id}] [{self.__class__.__name__}]")
        logger.setLevel(logging.DEBUG)

        stdout = logging.StreamHandler()
        logger.addHandler(stdout)

        stdout.setFormatter(logging.Formatter('%(asctime)s %(name)s %(message)s'))
        return logger

    def timeout(self):
        self.timeout_event.set()
        self.logger.debug('SIGTIMEOUT')

    def raise_for_timeout(self):
        if self.timeout_event.is_set():
            raise TimeoutException

    def is_timeout(self):
        return self.timeout_event.is_set()



import time
import logging
import traceback

from boastlabs.functions.retriable.execution.exceptions import RetryException, SleepException
from boastlabs.functions.retriable.execution.config import Status
from boastlabs.functions.retriable.execution.events.handling import EventHandling
from boastlabs.functions.retriable.execution.time import Timer
from boastlabs.functions.retriable.execution.state import State
from boastlabs.functions.retriable.execution.worker import TestWorker


class Function:
    def __init__(self, service_name, db, event_path, worker_class=TestWorker):

        self.service_name = service_name
        self.db = db
        self.event_path = event_path
        self.worker_class = worker_class

        self.event_ref = db.document(event_path)
        self.parent_ref = self.event_ref.parent.parent

        self.event_id = self.event_ref.id
        self.logger = self._init_logger()

    def _init_logger(self):
        logger = logging.getLogger(f"[{self.event_id}] [{self.__class__.__name__}]")
        logger.setLevel(logging.DEBUG)

        stdout = logging.StreamHandler()
        logger.addHandler(stdout)

        stdout.setFormatter(logging.Formatter('%(asctime)s %(name)s %(message)s'))
        return logger

    def try_run(self, timeout_seconds):
        start = time.time()

        # Init the timer
        timer = Timer(event_id=self.event_id)

        # Handles event and set the new state
        # Can raise ExecutionNotAllowed, DocumentNotFound, HandlerNotFound, RetryException
        event_handler = EventHandling(db=self.db, event_ref=self.event_ref)
        event_handler.handle()

        # Grab the state document, also the parent document
        # and figure out if we should run in "no op" mode or "real"
        # This is just for debug purposes
        state = State(service_name=self.service_name, db=self.db, job_ref=self.parent_ref)

        if state.etl_job_dict.get('worker_type', None) == 'dummy':
            worker = TestWorker(timer=timer, state=state, event_id=self.event_id)
        else:
            worker = self.worker_class(timer=timer, state=state, event_id=self.event_id)

        worker.start()

        # Wait TIMEOUT_MILLIS seconds and timeout 30 seconds early
        worker.join(timeout_seconds - 30)

        # Timeout here - from now on we have 30 seconds to finish threads
        timer.timeout()

        if worker.is_alive():
            self.logger.debug('# SIGTIMEOUT Sent')

        # Wait for thread to finish
        worker.join()

        self.logger.debug(f"# Worker joined with {state.status}")

        end = time.time()
        duration = end - start
        self.logger.debug(f'# Durations in seconds: {int(duration)}, (start={start}, end={end})')

        return state.status

    def run(self, timeout_seconds):
        self.logger.debug("# START")

        exit_status = None
        try:
            exit_status = self.try_run(timeout_seconds)
        except RetryException:
            # In the case that we could not handle the event due to "too much contention on these documents"
            # Then just try again the handler
            exit_status = Status.WAITING_RETRY
        except Exception:
            traceback.print_exc()
        finally:
            self.logger.debug("# END")

        # Raise exceptions here so GCP will retry our function
        if exit_status == Status.WAITING_RETRY:
            raise RetryException
        if exit_status == Status.WAITING_SLEEP:
            raise SleepException
        if exit_status == Status.FAILED:
            self.logger.error('# Function exited with Exception.')
        if exit_status == Status.DONE:
            self.logger.debug('# Function exited with success.')



import logging
from abc import abstractmethod

from google.cloud import firestore_v1
from google.cloud.firestore_v1 import DocumentReference

from boastlabs import __version__
from boastlabs.functions.retriable.execution.config import Status
from boastlabs.functions.retriable.execution.exceptions import ExecutionNotAllowedException, DocumentNotFoundException


class Event(object):

    def __init__(self, transaction: firestore_v1.Transaction, event_ref: DocumentReference):
        self.transaction = transaction
        self.event_ref = event_ref
        self.id = event_ref.id

        event_doc = self.event_ref.get(transaction=self.transaction)

        if not event_doc.exists:
            raise DocumentNotFoundException(doc_path=event_ref.path)

        self.data = event_doc.to_dict()

        self.event_type = self.data.get('event_type', None)
        self.api_version = self.data.get('api_version', __version__)
        self.event_handled = self.data.get('handled', False)

    def read(self):
        return self.event_ref.get(transaction=self.transaction).to_dict()

    def set_api_version(self):
        self.transaction.update(self.event_ref, {'api_version': __version__})

    def set_error(self, error):
        self.transaction.update(self.event_ref, {'error': error})

    def set_handled(self, handled):
        self.transaction.update(self.event_ref, {'handled': handled})

    def increase_invocation_count(self):
        self.transaction.update(self.event_ref, {'invocation_count': firestore_v1.transforms.Increment(1)})

    def increase_execution_count(self):
        self.transaction.update(self.event_ref, {'execution_count': firestore_v1.transforms.Increment(1)})

    def add_invocation(self, invocation):
        self.transaction.update(self.event_ref, {'invocations': firestore_v1.transforms.ArrayUnion([invocation])})


class TransactionalHandler(object):

    def __init__(self, event: Event):
        self.event = event
        self.transaction = self.event.transaction
        self.logger = self._init_logger()

    def _init_logger(self):
        logger = logging.getLogger(f"[{self.event.id}] [{self.__class__.__name__}]")
        logger.setLevel(logging.DEBUG)

        stdout = logging.StreamHandler()
        logger.addHandler(stdout)

        stdout.setFormatter(logging.Formatter('%(asctime)s %(name)s %(message)s'))
        return logger

    @abstractmethod
    def handle(self):
        raise NotImplementedError


class TriggerEventHandler(TransactionalHandler):

    def handle(self):
        # IMPORTANT
        # Reads first

        state_ref = self.event.event_ref.parent.parent
        state_dict = state_ref.get(transaction=self.transaction).to_dict()

        # IMPORTANT
        # Writes after
        old_status = state_dict.get('status', Status.NEW)
        new_status = old_status

        allow_execution = False

        if old_status in [Status.NEW, Status.WAITING_RETRY]:
            allow_execution = True
            new_status = Status.RUNNING
            self.transaction.update(state_ref, {'status': new_status})

        if old_status == Status.WAITING_SLEEP:
            allow_execution = True
            new_status = Status.SLEEP
            self.transaction.update(state_ref, {'status': new_status})

        invocation = {
            'old_status': old_status,
            'new_status': new_status,
            'allow_execution': allow_execution,
        }
        self.logger.debug(invocation)

        self.event.add_invocation(invocation)
        self.event.increase_invocation_count()
        self.transaction.update(state_ref, {'invocation_count': firestore_v1.transforms.Increment(1)})

        if allow_execution:
            self.event.increase_execution_count()
            self.transaction.update(state_ref, {'execution_count': firestore_v1.transforms.Increment(1)})
        else:
            raise ExecutionNotAllowedException(status=old_status)



import logging
import semantic_version

from google.cloud import firestore_v1
from google.cloud.firestore_v1 import DocumentReference

from boastlabs import __version__
from boastlabs.functions.retriable.execution.exceptions import (
    ExecutionNotAllowedException,
    HandlerNotFoundException,
    IncompatibleApiVersionException,
    RetryException)
from boastlabs.functions.retriable.execution.events.handler import Event, TriggerEventHandler
from boastlabs.functions.retriable.execution.config import Status


MAX_ATTEMPTS = 50


class EventHandling(object):

    def __init__(self, db, event_ref: DocumentReference):
        self.db = db
        self.event_ref = event_ref

        self.handlers = {}

        self.register_handler('trigger', TriggerEventHandler)

    def register_handler(self, event_type, handler):
        self.handlers[event_type] = handler

    def get_handler(self, event_type):
        return self.handlers.get(event_type, None)

    def handle(self):

        @firestore_v1.transactional
        def handle_in_transaction(transaction: firestore_v1.Transaction):
            # IMPORTANT
            # Read operations must be first

            event = Event(event_ref=self.event_ref, transaction=transaction)

            # Check api version compatibility
            if semantic_version.Version(event.api_version).major == semantic_version.Version(__version__).major:
                handler_class = self.get_handler(event.event_type)
                if handler_class:
                    handler = handler_class(event=event)
                    handler.handle()
                    event.set_api_version()
                    event.set_handled(True)
                else:
                    raise HandlerNotFoundException(event_type=event.event_type,
                                                   known_events=[e for e in self.handlers])
            else:
                raise IncompatibleApiVersionException(found_api_version=event.set_api_version(),
                                                      current_api_version=__version__)
        try:
            handle_in_transaction(transaction=firestore_v1.Transaction(client=self.db, max_attempts=MAX_ATTEMPTS))
        except ExecutionNotAllowedException as e:
            raise e
        except HandlerNotFoundException as e:
            raise e
        except IncompatibleApiVersionException as e:
            raise e
        except Exception:
            raise RetryException


class EventWorkDone(object):

    def __init__(self, db):
        self.db = db
        self.logger = self._init_logger()

    def _init_logger(self):
        logger = logging.getLogger(f"[{self.__class__.__name__}]")
        logger.setLevel(logging.DEBUG)

        stdout = logging.StreamHandler()
        logger.addHandler(stdout)

        stdout.setFormatter(logging.Formatter('%(asctime)s %(name)s %(message)s'))
        return logger

    def create(self,
               job_ref: DocumentReference,
               parent_ref: DocumentReference,
               service_name: str,
               service_status: str):

        @firestore_v1.transactional
        def handle_in_transaction(transaction: firestore_v1.Transaction):
            transaction.update(job_ref, {'active': False, 'status': Status.DONE})
            transaction.update(parent_ref, {'status': Status.DONE})

            event_ref = parent_ref.collection('events').document()

            transaction.set(event_ref, {
                'service_name': service_name,
                'service_status': service_status,
                'event_type': 'trigger'
            })

        handle_in_transaction(transaction=firestore_v1.Transaction(client=self.db, max_attempts=MAX_ATTEMPTS))



from abc import abstractmethod
import logging
from threading import Thread
from concurrent.futures import ThreadPoolExecutor, as_completed

from boastlabs.functions.retriable.execution.time import Timer


class TimedThread(Thread):

    def __init__(self, timer: Timer, event_id: str, name: str):
        Thread.__init__(self)

        self.timer = timer
        self.event_id = event_id

        self.setName(f"[{event_id}] [{name}]")
        self.logger = self._init_logger()

    def _init_logger(self):
        _logger = logging.getLogger(self.getName())
        _logger.setLevel(logging.DEBUG)

        stdout = logging.StreamHandler()
        _logger.addHandler(stdout)

        stdout.setFormatter(logging.Formatter('%(asctime)s %(name)s %(message)s'))
        return _logger

    def raise_for_timeout(self):
        if self.timer.is_timeout():
            self.logger.debug('# SIGTIMEOUT received, aborting.')
            self.timer.raise_for_timeout()

    @abstractmethod
    def work(self):
        raise NotImplementedError

    def run(self) -> None:
        self.logger.debug('# START')
        self.raise_for_timeout()

        self.work()
        self.logger.debug('# END')


class Executor(object):

    @classmethod
    def execute(cls, workers, max_workers):
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            for future in as_completed([executor.submit(worker.run) for worker in workers]):
                if future.exception():
                    raise future.exception()



from abc import abstractmethod
import random
import time

from boastlabs.functions.retriable.execution.exceptions import RetryException, SleepException
from boastlabs.functions.retriable.execution.threading import TimedThread
from boastlabs.functions.retriable.execution.state import State
from boastlabs.functions.retriable.execution.time import Timer
from boastlabs.functions.retriable.execution.config import Status


class Worker(TimedThread):

    def __init__(self, timer: Timer, event_id: str, state: State):
        TimedThread.__init__(self, timer=timer, event_id=event_id, name=self.__class__.__name__)

        self.state = state

    def get_sleep_duration(self):
        raise 0

    @abstractmethod
    def work(self):
        raise NotImplementedError

    def run(self):
        self.state.set_created_at()
        self.state.set_modified_at()

        if self.state.status == Status.SLEEP:
            self.state.set_status(Status.SLEEP)

            duration = self.get_sleep_duration()

            # Wait duration / can be interrupted by timeout

            self.logger.debug(f"# SLEEP START {duration}")
            self.timer.timeout_event.wait(duration)
            self.logger.debug(f"# SLEEP END")

            # Move to WAITING_RETRY state and resume execution on next call

            self.state.set_status(Status.WAITING_RETRY)

        if self.state.status == Status.RUNNING:
            self.logger.debug(f"# RUNNING")
            try:
                self.state.set_status(Status.RUNNING)
                self.work()
                self.state.set_status(Status.DONE)
            except RetryException:
                self.state.set_status(Status.WAITING_RETRY)
            except SleepException:
                self.state.set_status(Status.WAITING_SLEEP)
            except Exception as e:
                self.state.set_status(Status.FAILED, error=e)
                raise e


class TestWorker(Worker):

    def work(self):
        sleep = random.randint(0, 15)
        self.logger.debug(f"Sleep {sleep}")
        time.sleep(sleep)

        # if (random.randint(0, 9) == 1):
        #     raise Exception('Ingester Kind #1 exception')
        # if (random.randint(0, 9) == 2):
        #     raise Exception('Ingester Kind #2 exception')



class HandlerNotFoundException(Exception):

    def __init__(self, event_type, known_events):
        Exception.__init__(self, f'No value specified for field "event_type" or handler unknown'
                                 f' for value <{event_type}>. Known events: {known_events}')


class DocumentNotFoundException(Exception):

    def __init__(self, doc_path):
        Exception.__init__(self, f'No document found for path <{doc_path}>')


class IncompatibleApiVersionException(Exception):

    def __init__(self, found_api_version, current_api_version):
        Exception.__init__(self, f'Incompatible api version. Found <{found_api_version}>. '
                                 f'Current version: <{current_api_version}>')


class ExecutionNotAllowedException(Exception):

    def __init__(self, status):
        super().__init__(f'Execution Not Allowed! Status is already <{status}>.')


class SleepException(Exception):

    def __init__(self):
        super().__init__('Awaiting sleep!')


class RetryException(Exception):

    def __init__(self):
        super().__init__('Awaiting retry!')


class TimeoutException(RetryException):
    pass



from typing import List


class Progress(object):
    _id: str
    _overall: dict
    _items: List

    def __init__(self, id, progress=None):
        self._id = id

        if progress:
            self._overall = progress.get('overall')
            self._items = [Progress(id=item['id'], progress=item) for item in progress['items']]
        else:
            self._overall = {
                'ready': False,
                'page': 1,
                'count': 0,
            }
            self._items = []

    def __getitem__(self, item):
        items = [i for i in self._items if i._id == item]
        if len(items) == 0:
            item = Progress(id=item)
            self._items.append(item)
            return item
        if len(items) == 1:
            return items[0]
        return None

    def update(self, ready, count=None, page=None):
        self._overall['ready'] = ready

        if count:
            self._overall['count'] = count
        if page:
            self._overall['page'] = page

    def to_dict(self):
        return {
            'id': self._id,
            'overall': {
                'ready': self.is_ready(),
                'page': self.get_page(),
                'count': self.count(),
                'children_count': self.children_count(),
                'children_ready_count': self.children_ready_count()
            },
            'items': [item.to_dict() for item in self._items]
        }

    def is_ready(self):
        # Leaf node, return ready state
        if len(self._items) == 0:
            return self._overall['ready']
        else:
            # Compute downwards the ready state
            is_ready = True
            for item in self._items:
                is_ready = is_ready and item.is_ready()
            return is_ready

    def get_page(self):
        return self._overall['page']

    # Only for UI
    def count(self):
        return self._overall['count']

    # Only for UI
    def children_count(self):
        return len(self._items)

    # Only for UI
    def children_ready_count(self):
        return len([i for i in self._items if i.is_ready()])
